#! /usr/bin/env python

import networkx as nx

import os
import glob
import sys

import traceback

#import numpy as np

import xml.etree.ElementTree as ET

import cPickle

import urllib

import csv
import re
import time

import pprint as pp

import threading
from Queue import Queue

import urllib2

#TODO: move cache and country handling into script


#ToDO: fix logging handling now in seperate script
import logging
import logging.handlers


#TODO: make logging passed in for Class for geocode

#TODO: redo caching based on decorator using tuple of arguments passed in
# look at tuple of arguments/url?
class FileGetter(threading.Thread):
    # based on http://www.artfulcode.net/articles/multi-threading-python/
    def __init__(self, node, url ):
        self.url = url
        self.result = None
        self.node = node
        self.has_exceeded_count = False
        self.logger = logging.getLogger("geocoder")
        threading.Thread.__init__(self)
        

    def get_result(self):
        return self.result

    def get_node(self):
        return self.node

    def get_url(self):
        return self.url

    def exceeded_count(self):
        return self.has_exceeded_count

    def process_result(self, web_result):
        results = []
        for entry in web_result:
            place_data = {}
            for attr in [ "name", "countryCode", "lat", "lng", "adminName1",
                         "population", "countryName", "geonameId"]:
                try:
                    value = entry.find(attr)
                    if value is not None:
                        value = value.text
                        value = value.encode('utf-8')
                except AttributeError, e:
                    # Problem with attribute, use default
                    self.logger.debug("AttributeError %s" % e)
                    value = ""
                place_data[attr] = value
            results.append(place_data)
        # return only the entry for single results
        if len(results) == 1:
            results = results[0]
        elif len(results) == 0:
            results = None

        return results

    def run(self):
        overloaded_message = ("free servers are currently "
                                "overloaded with requests" )
        exceeded_message = ("Please throttle your requests "
                            "or use the commercial service")
        for i in range(10):
            try:
                self.logger.debug("Fetching %s"% self.url)
                geocode_url = urllib2.urlopen(self.url)
                geocode_result = geocode_url.read()
                geocode_url.close()
                if overloaded_message in geocode_result:
                    self.logger.debug("Overloaded server, "
                                 "trying again in 0.5 seconds")
                    time.sleep(0.5)
                elif exceeded_message in geocode_result:
                    self.logger.warn("Exceeded limit for geonames server")
                    #TODO: make sure don't cache this result
                    self.has_exceeded_count = True
                    return None
                else:
                    # Result is fine
                    break
            except urllib2.HTTPError, e:
                self.logger.debug(e)
                time.sleep(0.5)
    
        dom = ET.XML(geocode_result)
        #TODO: look at more programatic way - eg look at first tag
        if "http://api.geonames.org/get" in self.url:
            # geoname_id result
            # Only single result returned, <geoname>...</geoname>
            # Pass this to the iteration as a list to be processed as per normal
            geonames = [dom]
        else:
            geonames =  dom.findall("geoname")

        self.result = self.process_result(geonames)

class Geocoder(object):

    def __init__(self, geonames_username, skip_cache):
        self.skip_cache = skip_cache
        if not geonames_username:
            #TODO: raise error
            print "Username must be specified for GeoNames"
            return
        self.geonames_username = geonames_username


        #TODO: work out best way to put in function but make globally accessible
        self.logger = logging.getLogger("geocoder")
        self.logger.setLevel(logging.DEBUG)
        formatter = logging.Formatter('%(levelname)-6s %(message)s')
        ch = logging.StreamHandler()
        ch.setLevel(logging.INFO)
        ch.setFormatter(formatter)
        logging.getLogger('').addHandler(ch)
        log_dir = "."
        LOG_FILENAME =  os.path.join(log_dir, "geocoder.log")
        LOG_SIZE = 2097152 # 2 MB
        fh = logging.handlers.RotatingFileHandler(
            LOG_FILENAME, maxBytes=LOG_SIZE, backupCount=5)
        fh.setLevel(logging.DEBUG)
        formatter = logging.Formatter("%(asctime)s %(levelname)s %(message)s")
        fh.setFormatter(formatter)

        self.web_cache = {}
        self.cache_file = 'cache.data'
        if (os.path.isfile(self.cache_file)):
            f_web_cache = open(self.cache_file, 'rb')
            self.web_cache = cPickle.load(f_web_cache)
            f_web_cache.close()

        self.last_written_cache_length = len(self.web_cache)

        self.country_info = {}
        country_info_file = 'country_info.data'
        if (os.path.isfile(country_info_file)):
            self.country_info = cPickle.load(open(country_info_file, 'rb'))
        else:
            self.logger.debug("Downloading country information "
                              "file from GeoNames")
            #TODO: rename this to be geocode_data as csv_data is used elsewhere
            csv_data = urllib2.urlopen("http://download.geonames.org/export/dump/countryInfo.txt")
            fieldnames = ["ISO", "ISO3", "ISONumeric", "fips", "Country", "Capital",
                        "Area", "Population", "Continent", "tld", "CurrencyCode",
                        "CurrencyName", "Phone", "PostalCodeFormat",
                        "PostalCodeRegex", "Languages", "geonameid", "Neighbors",
                        "EquivalentFipsCode"]
            csv_reader = csv.DictReader( csv_data, fieldnames = fieldnames,
                                        delimiter="\t")
            for line in csv_reader:
                c_iso = line['ISO']
                # Skip comment lines
                if c_iso[0] != "#":
                    self.country_info[c_iso] = {
                            'ISO3': line['ISO3'],
                            'Country': line['Country'],
                            'fips': line['fips'],
                            'Continent': line['Continent'],
                            'Neighbors': line['Neighbors'],
                            }
            csv_data.close()
            # And store pickled result
            self.logger.debug("Saving processed country information file")
            f = open(country_info_file, 'wb')
            cPickle.dump(self.country_info, f, -1)
            f.close()

    def save_cache(self):
        # Only write if changed length since last time
        if len(self.web_cache) != self.last_written_cache_length:
            f_web_cache = open(self.cache_file, 'wb')
            cPickle.dump(self.web_cache, f_web_cache, -1)
            f_web_cache.close()
            self.last_written_cache_length = len(self.web_cache) 

    def city_cleanup(self, city):
        # cleans up various data from city name
        city = city.rstrip("123456789?")

        # Slices are faster than regular expressions
        # Expand eg Ft Lauderdale to Fort Lauderdale
        if city[:3].lower() == "ft ":
            city = "Fort " + city[3:]

        if city[:3].lower() == "pt ":
            city = "Port " + city[3:]

        if city[:3].lower() == "mt ":
            city = "Mount " + city[3:]

        if city[:3].lower() == "st ":
            city = "Saint " + city[3:]

        # and for dot instead of space
        if city[:3].lower() == "ft.":
            city = "Fort" + city[3:]

        if city[:3].lower() == "pt.":
            city = "Port" + city[3:]

        if city[:3].lower() == "mt.":
            city = "Mount" + city[3:]

        if city[:3].lower() == "st.":
            city = "Saint" + city[3:]

        # Ending in Mt -> Mountain
        if city[-3:].lower() == " mt":
            city = city[:-3] + " Mountain"

        # Datacentre as DC
        if city[-3:].lower() == " dc":
            if city.lower() != "washington dc":
                city = city[:-3]

        # U eg for University, both at start and end
        if city[:2].lower() == "u ":
            city = city[2:]
        if city[-2:].lower() == " u":
            city = city[:-2]


        return city

    def country_key(self, country):
        for key, val in self.country_info.iteritems():
            if ( val['Country'] == country
                    or val['ISO3'] == country
                    or val['fips'] == country):
                return key

    def countries_for_continent(self, continent):
        results = []
        # Lookup countries in this continent
        #TODO: use general string type checking str and unicode inherit from
        if isinstance(continent, str) or isinstance(continent, unicode):
            results = [key for (key, val) in self.country_info.iteritems()
                    if val['Continent'] == continent]

        elif isinstance(continent, list):
            results = [key for (key, val) in self.country_info.iteritems()
                    if val['Continent'] in continent]

        return results

    def country_code(self, graph):
        name = None
        code = None
        if 'GeoExtent' in graph.graph:
            geoextent = graph.graph['GeoExtent']
            geolocation = graph.graph['GeoLocation']
            continent_codes = {
                'North America':    'NA',
                'Europe':           'EU',
                # Mexico, Honduras, etc are in North America in Geonames
                'Latin America':      ['SA', 'NA'],
                'Africa':           'AF',
                'Middle East':      'AF',
                # Asia-Pacific from zoo is two continents in Geonames
                'Asia-Pacific':      ['AS', 'OC'],
            }

            if geoextent == 'Region':
                # Extract the last bit of the comma seperated region as country
                # eg Washington State, USA -> extract 'USA'
            #TODO: look at doing for eg South Australia, Victoria, Australia
                fragments = geolocation.split(",")
                #TODO: use slice notation here
                name = fragments[len(fragments)-1].strip()
            elif geoextent == 'Country':
                name = geolocation
            elif geoextent == 'Country+':
                # convert into array and remove spaces
                name = [x.strip() for x in geolocation.split(",") ]
                code = []
                # find continents and countries present (can have both, eg USA,
                # Europe)
                continents = [continent_codes[con] for con in name if con in
                        continent_codes]
                for cont in continents:
                    code += self.countries_for_continent(cont)
                if 'Latin America' in name:
                    if 'US' in code:
                        code.remove('US')
                    if 'CA' in code:
                        code.remove('CA')
                # And key for countries
                code += [self.country_key(cou) for cou in name if cou not
                        in continent_codes]

            elif geoextent == 'Continent':
                code = []
                # get the countries in this continent
                if geolocation in continent_codes:
                    continent = continent_codes[geolocation]
                    # Fetch list of countries in this continent:
                    code = self.countries_for_continent(continent)
                    if geolocation == 'Latin America':
                        if 'US' in code:
                            code.remove('US')
                        if 'CA' in code:
                            code.remove('CA')
                else:
                    self.logger.warn("Continent %s not"
                                "in list of continents"%geolocation)
            elif geoextent == 'Continent+':
                code = []
                # convert into array and remove spaces
                name = [x.strip() for x in geolocation.split(",") ]
                # Find continent code
                continents = [continent_codes[con] for con in name if con in
                        continent_codes]
                # Find countries for continent codes
                for cont in continents:
                    code += self.countries_for_continent(cont)

                # If Latin-America, remove USA and Canada from the North America
                # join
                if 'Latin America' in name:
                    if 'US' in code:
                        code.remove('US')
                    if 'CA' in code:
                        code.remove('CA')
        if name and not code:
            # Country has been set, but no country code, try and look up code
            if isinstance(name, str) or isinstance(name, unicode):
                # Just look up the name
                code = self.country_key(name)
            if isinstance(name, list):
                # Multiple countries, eg from Country+
                code = []
                for cou in name:
                    code.append(self.country_key(cou))

        return code

    def geocode(self, graph):
        # How fuzzy we accept results to be [0,1], lower = accept 'fuzzier' answers
        fuzzy = 0.8

        c_code = self.country_code(graph)

        if isinstance(c_code, list):
            # Ensure same order for URL even if csv etc changes
            c_code = sorted(c_code)

        def create_url(node, iteration):
            # iteration is the pass, first try with query, then with different
            # country, etc
            params = {
                'maxRows': 1,
            }

            # See if place has geocode id manually set
            if 'geocode_id' in graph.node[node]:
                # use this instead of country code
                params['geonameId'] = graph.node[node]['geocode_id']
                params['style'] = "SHORT"
                params = urllib.urlencode(params)
                url = "http://api.geonames.org/get?{0}".format(params)
                # Different URL style, form here and return
                return url

            # See if place is specified as a country
            if ('geocode_country' in graph.node[node]):
                # search on this instead
                params['q'] = graph.node[node]['geocode_country']
                # Feature class for country, region, etc
                params['featureClass'] = 'A'
                params = urllib.urlencode(params)
                url = "http://api.geonames.org/search?{0}".format(params)
                # Different URL style, form here and return
                return url

            if 'geocode_name' in graph.node[node]:
                # Manually set city to use
                city = graph.node[node]['geocode_name']
            else:
                city = graph.node[node]['label']

            city = self.city_cleanup(city)

            #tODO: rename city to place

            if i == 0:
                params['featureClass'] = 'P'
                # Primary search query
                if 'geocode_extent' in graph.node[node]:
                    # use this instead of country code
                    if graph.node[node]['geocode_extent'] == "*":
                        # Wildcard, don't append extra, but don't use country
                        # code
                        params['q'] = city
                    else:
                        city_extent = city + ' ' + graph.node[node]['geocode_extent']
                        params['q'] = city_extent
                elif 'geocode_append' in graph.node[node]:
                    # use this instead of country code
                    city_appended = city + ' ' + graph.node[node]['geocode_append']
                    params['q'] = city_appended
                    if isinstance(c_code, list):
                        params['country'] = ",".join(c_code)
                    elif c_code == None:
                        # Don't lookup with &country=None
                        pass
                    else:
                        params['country'] = c_code
                elif ('GeoExtent' in graph.graph
                    and graph.graph['GeoExtent'] == 'Region'):
                    city_extent = city + ' ' + graph.graph['GeoLocation']
                    params['q'] = city_extent
                else:
                    params['q'] = city
                    if isinstance(c_code, list):
                        params['country'] = ",".join(c_code)
                    elif c_code == None:
                        # Don't lookup with &country=None
                        pass
                    else:
                        params['country'] = c_code
            elif i == 1:
                params['featureClass'] = 'P'
                # Try fuzzy match
                params['q'] = city
                params['fuzzy'] = fuzzy
                if isinstance(c_code, list):
                    params['country'] = ",".join(c_code)
                elif c_code == None:
                    # Don't lookup with &country=None
                    pass
                else:
                    params['country'] = c_code
            elif i == 2:
                # try for other feature classes - eg hill, area, etc
                params['q'] = city
                if isinstance(c_code, list):
                    params['country'] = ",".join(c_code)
                elif c_code == None:
                    # Don't lookup with &country=None
                    pass
                else:
                    params['country'] = c_code

            params = urllib.urlencode(params)
            url = "http://api.geonames.org/search?{0}".format(params)
            return url

        #TODO: cut out any University 1, Institution etc

        lookup_places = []

        for node, nodedata in graph.nodes(data=True):
            city = graph.node[node]['label']
            # Some nodes are skipped
            if 'skip_geocode' in nodedata and nodedata['skip_geocode'] == True:
                continue
            if 'Latitude' in nodedata and 'Longitude' in nodedata:
                self.logger.warn("Lat/Lon set for %s, skipping" % city)
                continue
            elif not city:
                # Blank city
                continue
            elif (city == '' 
                  or 'hyperedge' in nodedata and nodedata['hyperedge'] == True):
                # blank city
                continue
            elif ((len(city) <= 4 and city.isupper())
                and
                ('geocode_name' not in nodedata and 'geocode_id' not in nodedata
                and 'geocode_country' not in nodedata)):
                # Skip short capitalized names, as most likely an abbreviation
                # unless manually set geocode place or id
                continue
            elif ((len(city) <= 4 and city.isdigit())
                and
                ('geocode_name' not in nodedata and 'geocode_id' not in nodedata
                and 'geocode_country' not in nodedata)):
                # Skip numeric names unless manually set geocode place or id
                continue
            elif city == "?":
                continue
            elif ''.join([s for s in city if s.isalpha()]) in ["University",
                                                            "Institution"]:
                # Skip if only "Institution" or "University" as these are not
                # location specific, and may lead to spurious results
                # The first bit of code keeps only letters (ie remove numbers)
                # and is faster than using a regex
                continue
            elif nodedata['Internal'] == False:
                # Skip ext nodes as likely to be networks rather than places
                continue
            else:
                lookup_places.append(node)

        # Produced and Consumer function for thread
        def producer(q, urls):
            for node, url in urls.items():
                # Add the Geonames Username here before querying URL
                # if put in earlier then would complicate cache handling
                # It is only part of the way geonames is queried, not part of the
                # query
                url = "%s&username=%s"%(url, self.geonames_username)
                thread = FileGetter(node, url)
                thread.start()
                q.put(thread, True)

        def consumer(q, total_urls, web_result):
            while (len(web_result) < total_urls):
                thread = q.get(True)
                thread.join()
                result = thread.get_result()
                if not result and thread.exceeded_count():
                    # Exceeded tries, don't store result
                    pass
                else:
                    # either got result, or got no result, store either
                    node_id = thread.get_node()
                    web_result[node_id] = result

        geocoded = {}

        #TODO: need better handling for exceeded warning

        for i in range(3):
            # Find nodes that still need to be looked up
            need_geocode = []
            for n in lookup_places:
                if (n not in geocoded) or (n in geocoded and geocoded[n] == None):
                    need_geocode.append(n)

            # Generate lookup URLs
            urls_to_fetch = {}
            if len(urls_to_fetch) > 0:
                self.logger.debug("Lookup iteration %s"%i)
            for node in need_geocode:
                url = create_url(node, i)
                use_from_cache = False
                if ('skip_cache' in graph.node[node] and
                    graph.node[node]['skip_cache'] == True):
                    use_from_cache = False
                elif url in self.web_cache:
                    use_from_cache = True

                if self.skip_cache:
                    # User specified to not use cache
                    use_from_cache = False

                if use_from_cache:
                    # Already cached, no need to look up
                    geocoded[node] = self.web_cache[url]
                else:
                    # Need to lookup
                    urls_to_fetch[node] = url
                    
            # Now lookup the Urls
            web_result = {}
            #TODO: make queue parameter for max web threads at once
            q = Queue(10)
            prod_thread = threading.Thread(target=producer,
                                        args=(q, urls_to_fetch))
            cons_thread = threading.Thread(target=consumer,
                                        args=(q, len(urls_to_fetch), 
                                              web_result))
            prod_thread.start()
            cons_thread.start()
            prod_thread.join()
            cons_thread.join()

            # extract and cache results fetched from web
            for node, result in web_result.items():
                # Find the URL used for this node
                url = urls_to_fetch[node]
                # and store
                self.web_cache[url] = result
                # and store result
                geocoded[node] = result

        for node, data in geocoded.items():
            if data:
                graph.node[node]['Latitude'] =  float(data['lat'])
                graph.node[node]['Longitude'] = float(data['lng'])
                graph.node[node]['Country'] = data['countryName']
            else:
                self.logger.info( "No geocode match for " + 
                            graph.node[node]['label'])

        # and remove any geoname_id or geocode_name set now complete geocoding
        for node, data in graph.nodes(data=True):
            if 'geoname_id' in data:
                del graph.node[node]['geoname_id']
            if 'geocode_name' in data:
                del graph.node[node]['geocode_name']
            if 'geocode_extent' in data:
                del graph.node[node]['geocode_extent']

        return graph

def main():
    import optparse
    opt = optparse.OptionParser()
    opt.add_option('--file', '-f', help="Load data from FILE")
    opt.add_option('--directory', '-d', help="process directory")
    opt.add_option('--geonames_username', '-u', 
                default=False,
                help="Username for Geonames web service")
    opt.add_option('--skip_cache',  action="store_true",
                default=False,
                help="Don't use cache for geocoding")

    options = opt.parse_args()[0]

    if not options.geonames_username:
        print("Please enter a Geonames username to use geocoding")
        sys.exit(0)

    network_files = []
    if options.file:
        network_files.append(options.file)

    if options.directory:
        network_files = glob.glob(options.directory + "*.gml")

    if len(network_files) == 0:
        print "No files found. Please specify -f file or -d directory"
        sys.exit(0)

    if options.directory:
        path = options.directory
    elif options.file:
        path, filename = os.path.split(options.file)
        path = os.path.abspath(path)

    output_path = path + os.sep + "geocoded" 
    if not os.path.isdir(output_path):
        os.mkdir(output_path)
    print("Saving to folder: %s" % output_path)

    # and create cache directory for pickle files
    pickle_dir = path + os.sep + "cache"       
    if not os.path.isdir(pickle_dir):
        os.mkdir(pickle_dir)
    
    geocoder = Geocoder(options.geonames_username, options.skip_cache)
    
    for net_file in network_files:
        path, filename = os.path.split(net_file)
        network_name = os.path.splitext(filename)[0]
        print "Reading: %s" % network_name
        #TODO: add cache support
        pickle_file = "%s%s%s.pickle"% (pickle_dir, os.sep, network_name)
        if (os.path.isfile(pickle_file) and
            os.stat(net_file).st_mtime < os.stat(pickle_file).st_mtime):
            # Pickle file exists, and source_file is older
            graph = nx.read_gpickle(pickle_file)
        else:
            # No pickle file, or is outdated
            graph = nx.read_gml(net_file)
            nx.write_gpickle(graph, pickle_file)


        graph = geocoder.geocode(graph)
        geocoder.save_cache()
        gml_file =  "%s%s%s.gml"%s(output_path, os.sep, network_name)
        nx.write_gml(graph, gml_file)
        print("Wrote to %s" % gml_file)
   
if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        pass



#! /usr/bin/env python

import networkx as nx

import os
import glob
import sys
from time import strftime

import optparse
import csv

import pprint

from mako.lookup import TemplateLookup    
import zipfile
import cPickle
from collections import defaultdict


def continent_breakdown(geo_locations_by_extent):
    retval = {'GlCPlus': [], 'AP': [], 'Eur': [], 'NA/LA': []}

    country_info = {}
    country_info_file = 'country_info.data'
    if (os.path.isfile(country_info_file)):
        country_info = cPickle.load(open(country_info_file, 'rb'))

    # Mapping of country to continent
    country_continents = dict( (d['Country'], d['Continent']) for d in
                              country_info.values())
    # unique continent codes

    for values in geo_locations_by_extent['Global'].values():
        retval['GlCPlus'] += values
    for values in geo_locations_by_extent['Continent+'].values():
        retval['GlCPlus'] += values
    # Remove so can see leftovers at end
    del geo_locations_by_extent['Global']
    del geo_locations_by_extent['Continent+']

    # Now by continent
    cont_mapping = {'Asia-Pacific': 'AP', 'Europe': 'Eur', 
                    'North America': 'NA/LA', 'Latin America': 'NA/LA'}
    for key, val in cont_mapping.items():
        retval[val] += geo_locations_by_extent['Continent'][key]
        del geo_locations_by_extent['Continent'][key]

    # Now by country
    # Continents to group countries by
    # then convert these into groups for retval
    by_continent = {'AF': [], 'NA': [], 'OC': [], 'AN': [], 'AS': [], 'EU': [],
                    'SA': []}
    # Append manual country lookups
    country_continents['US'] = "NA"
    country_continents['Palestine'] = "AF"
    country_continents['Northern Ireland'] = "EU"
    country_continents['USA'] = "NA"
    country_continents['UK'] = "EU"

    for country, data in geo_locations_by_extent['Country'].items():
        if country in country_continents:
            # Have continent for this country
            cont = country_continents[country]
            by_continent[cont] += data
            del geo_locations_by_extent['Country'][country]
        else:
            print "no continent for %s" % country


    for countries, data in geo_locations_by_extent['Country+'].items():
        # For country+ we take the first country to find the continent
        country = countries.split(",")[0]
        if country in country_continents:
            # Have continent for this country
            cont = country_continents[country]
            by_continent[cont] += data
            del geo_locations_by_extent['Country+'][countries]
        else:
            print "no continent for country+ %s" % countries


    # Region
    for region, data in geo_locations_by_extent['Region'].items():
        # Region generally in format "region, country" so try split on comma
        country = region.split(",")[1].strip()
        if country in country_continents:
            cont = country_continents[country]
            by_continent[cont] += data
            del geo_locations_by_extent['Region'][region]
        else:
            print "no continent for region %s with country of %s" % (region,
                                                                     country)

    # same for metro
    for metro, data in geo_locations_by_extent['Metro'].items():
        # metro generally in format "metro, country" so try split on comma
        country = metro.split(",")[1].strip()
        if country in country_continents:
            cont = country_continents[country]
            by_continent[cont] += data
            del geo_locations_by_extent['Metro'][metro]
        else:
            print "no continent for metro %s with country of %s" % (metro,
                                                                    country)

    # And now map these back
    # Ignore Africa and Antarctica
    cont_mapping = {'NA': 'NA/LA', # North America -> North/Latin America
                    'OC': 'AP',     #Oceania -> Asia pac
                    'AS': 'AP',     # Asia -> asia pac
                    'EU': 'Eur',
                    'SA': 'NA/LA'   # South America -> North/Latin America
                   }
    for key, val in cont_mapping.items():
        retval[val] += by_continent[key]
        del by_continent[key]

    return retval

def matlab_keys(summary_data, output_path, counts_only=False):
    matlab_keys = {}

    keys_to_use = ["Classification", "Type", "Layer", "GeoExtent",
                    "GeoLocation", "Developed", "Dataset", "NetworkData",
                   "Provenance", "Access", "Backbone", "Customer", "Transit",
                   "Testbed", "IX", "NetworkDate"]

    # Index of this network in the directory listing. Note: starting at 1
    network_index = 1
    file_index = {}
    geo_extents = defaultdict(list)
    geo_locations_by_extent = defaultdict(dict)
    
    for network, data in sorted(summary_data.items()):

        file_index[network_index] = network

        geo_extent = data['GeoExtent']
        geo_location = data['GeoLocation']
        geo_extents[geo_extent].append(network_index)
        geo_locations_by_extent[geo_extent].setdefault(geo_location,
                                            []).append(network_index)

        # look at properties of this network 
        for key, val in data.items():
        
            if key not in keys_to_use:
                # Not interested in this key (eg source, version, etc), skip
                continue

            if key not in matlab_keys:
                matlab_keys[key] = {}
            
            if val not in matlab_keys[key]:
                matlab_keys[key][val] = []

            matlab_keys[key][val].append(network_index) 


        network_index += 1

    # Custom keys for Hung for JSAC
    # breakdown by country/region/multi
    # Store for printing
    matlab_keys['cou_reg_mul'] = {}
    matlab_keys['cou_reg_mul']['Country'] = (geo_extents['Country'] +
                                             geo_extents['Country+'])
    matlab_keys['cou_reg_mul']['Region'] = (geo_extents['Metro'] +
                                            geo_extents['Region'])
    matlab_keys['cou_reg_mul']['Multi'] = (geo_extents['Continent'] + 
                                           geo_extents['Continent+'] + 
                                           geo_extents['Global'])

    c_breakdown = continent_breakdown(geo_locations_by_extent)
    matlab_keys['continent_breakdown'] = c_breakdown
    # Sanity check
    #for continent, values in c_breakdown.items():
    #    print continent
    #    print "; ".join([summary_data[file_index[network_index]]['GeoLocation'] for
    #                    network_index in values])

    f_matlab_keys = open("{0}/matlab_keys.txt".format(output_path), "w")
    
    if not counts_only:
        f_matlab_keys.write("-----------\nFile Index\n")
        for key, val in sorted(file_index.items()):
            f_matlab_keys.write("{0} {1}\n".format(key, val))
    # Also print out in a list for Matlab purposes so can select on index
    f_matlab_keys.write("\n['" + "', '".join(file_index.values()) + "']")
    for heading, data in matlab_keys.items():
        f_matlab_keys.write("\n-----------------\n" + heading + "\n")
        # do keys with names at end so can copy paste first bit into matlab
        keys_with_names = '\nNamed:\n'
        keys_with_counts = '\nCounts:\n'
        for key, values in sorted(data.items()):
            if key == '':
                key = "No value"
            # Also list with names for double checking
            if not counts_only:
                f_matlab_keys.write("{0}\t{1}\n".format(key, sorted(values)))
                values_named = sorted([file_index[x] for x in values])
                keys_with_names += "{0}\t[{1}]\n".format(key, 
                                                         ','.join(values_named))
            keys_with_counts += "{0}\t{1}\n".format(key, len(values))
        if not counts_only:
            f_matlab_keys.write(keys_with_names)
        f_matlab_keys.write(keys_with_counts)
    f_matlab_keys.close()


def main():
    #toDO: move parser out of main
    opt = optparse.OptionParser()
    opt.add_option('--file', '-f', help="Load data from FILE")
    opt.add_option('--directory', '-d', help="process directory")
    opt.add_option('--output_dir', '-o', help="process directory")
    opt.add_option('--map_dir', help="geoplot map input dir for gallery")


    # Graph options
    opt.add_option('--single_edge', action="store_true",
                   default=False, help="Reduce multi edges to single edge")
    opt.add_option('--internal_only', action="store_true",
                   default=False, help="Exclude external nodes")
    opt.add_option('--matlab_keys', action="store_true",
                   default=False, help="Generate list of keys for Matlab")
    opt.add_option('--counts_only', action="store_true",
                   default=False, help="Only record counts for Matlab keys")
    # Output options
    opt.add_option('--matlab', action="store_true", default=False,
                   help="Write Matlab adjacency list Format")
    opt.add_option('--gml',  action="store_true", default=False, 
                   help="Write GML Format")
    opt.add_option('--graphml', action="store_true", default=False,
                   help="Write Graphml Format")
    opt.add_option('--dot', action="store_true", default=False, 
                   help="Write GraphViz dot Format")
    opt.add_option('--gexf', action="store_true", default=False,
                   help="Write GEXF Format")
    opt.add_option('--plot', action="store_true", default=False,
                   help="Plot graph")
    opt.add_option('--gallery', action="store_true", default=False,
                   help="Render gallery HTML")

    opt.add_option('--summary_csv', action="store_true", default=False,
                   help="Generate summary data in CSV format")


    opt.add_option('--pickle', action="store_true", default=False,
                   help="Write NetworkX Pickle Format")
    opt.add_option('--archive', action="store_true", default=False,
                   help="Archive converted files")
    opt.add_option('--html', action="store_true", default=False, 
                   help="HTML summary")
    opt.add_option('--skip_type',  help="List of types to skip")
    options = opt.parse_args()[0]

    network_files = []
    if options.file:
        network_files.append(options.file)

    if options.directory:
        network_files = glob.glob(options.directory + "*.gml")

    if len(network_files) == 0:
        print "No files found. Please specify -f file or -d directory"
        sys.exit(0)

    if options.directory:
        path = options.directory
    elif options.file:
        path, filename = os.path.split(options.file)
        path = os.path.abspath(path)

    if options.output_dir:
        output_path = options.output_dir 
    else:
        output_path = path + os.sep + "converted"
    
    if not os.path.isdir(output_path):
        os.mkdir(output_path)  


    pickle_dir = path + os.sep + "cache"
    if not os.path.isdir(pickle_dir):
        os.mkdir(pickle_dir)

    summary_data = {}

    archive_file =  None
    if options.archive:
        archive_filelist = []
        archive_file = "archive.zip"

    #These headings are used in HTML template, if not present then set to "" 
    metadata_headings = ["Network", "GeoExtent", "GeoLocation", 
                         "Type", "Classification", "LastAccess",
                        "Source", "Layer", "DateObtained",
                        "NetworkDate", "Note"]
    for net_file in network_files:

        # Extract name of network from file path
        path, filename = os.path.split(net_file)
        network_name = os.path.splitext(filename)[0]

        print "Converting: {0}".format(network_name)

        #if network_files.index(net_file) > 20:
        #    break

        pickle_file = "{0}/{1}.pickle".format(pickle_dir, network_name)
        if (os.path.isfile(pickle_file) and
            os.stat(net_file).st_mtime < os.stat(pickle_file).st_mtime):
            # Pickle file exists, and source_file is older
            graph = nx.read_gpickle(pickle_file)
        else:
            # No pickle file, or is outdated
            graph = nx.read_gml(net_file)
            nx.write_gpickle(graph, pickle_file)


        graph = graph.to_undirected()
        graph = nx.MultiGraph(graph)

        if options.skip_type:
            skip_type = options.skip_type
            print "Skipping {0}".format(skip_type)
            # Search for nodes with this type set
            for node, data in graph.nodes(data=True):
                if 'type' in data and data['type'] == skip_type:
                    graph.remove_node(node)

        if (options.html or options.matlab_keys or options.gallery or
            options.summary_csv):
            summary_data[network_name] = {} 

            for key, val in graph.graph.items():
                # And also store in html data
                summary_data[network_name][key] = val
            # also some graph statistics
            summary_data[network_name]['Nodes'] = graph.number_of_nodes()
            summary_data[network_name]['Edges'] = graph.number_of_edges()

        #*********************************
        #Graphs - create different graphs for writing
        #*********************************
        if options.single_edge:
            # Graph with max one edge between any node pair
            graph = nx.Graph(graph)

        #TODO Update this to be true/false once netx fixes boolean quote bug
        if options.internal_only:
            external_nodes = [ n for n in graph.nodes()
                              if ('Internal' in graph.node[n] and 
                              graph.node[n]['Internal'] == 0)]
            graph.remove_nodes_from(external_nodes)
        # ************

        out_file = "{0}/{1}".format(output_path, network_name)
        # Append any custom options to filename
        if options.skip_type:
            #TODO Make sure only alphanumeric (safe file name) chars in filename
            out_file += "_skip_" + options.skip_type.lower()
        if options.single_edge:
            out_file += "_se" 
        if options.internal_only:
            out_file += "_int"

        #*********************************
        #OUTPUT - Write the graphs to files
        #*********************************
        if options.gml:
            #Add quotes
            gml_file =  out_file + ".gml"
            # workaround for bug where gml writer removes label from node
            #TODO: see if this is needed
            graph2 = graph.copy()
            nx.write_gml(graph2, gml_file)
            if options.archive:
                archive_filelist.append(gml_file)
   
        if options.graphml:
            graphml_file =  out_file + ".graphml"
            nx.write_graphml(graph, graphml_file)
            if options.archive:
                archive_filelist.append(graphml_file) 
     
        if options.dot:
            dot_file =  out_file + ".dot"
            nx.write_dot(graph, dot_file) 
            if options.archive:
                archive_filelist.append(dot_file)
        
        if options.gexf:
            gexf_file =  out_file + ".gexf"
            nx.write_gexf(graph, gexf_file) 
            if options.archive:
                archive_filelist.append(gexf_file)

        if options.pickle:
            pickle_file =  out_file + ".pickle"
            nx.write_gpickle(graph, pickle_file) 
            if options.archive:
                archive_filelist.append(pickle_file)

        #TODO: merge options.plot with geoplot
        # make geoplot part of this

        if options.matlab:
            # Write as a sparse matrix
            mat_file =  out_file + ".mat"
            mat_fh = open(out_file + ".txt", "w")
            if options.archive:
                archive_filelist.append(mat_file)
            sparse_matrix = nx.to_numpy_matrix(graph)

            # Write out manually (numpy.tofile() writes as all on one line)
            for line in sparse_matrix.tolist():
                # Convert list to a string format numbers as integers not floats
                line = [ str(int(x)) for x in line]
                mat_fh.write( " ".join(line) )
                mat_fh.write("\n")
            mat_fh.close()

    if options.matlab_keys:
        matlab_keys(summary_data, output_path, options.counts_only)

    if options.summary_csv:
        network_list = [data['Network'] for data in summary_data.values()]
        fieldnames = ['Name', 'Network', 'Date', 'Type', 'Layer',
                      'Nodes', 'Edges',
                      'Developed', 'GeoExtent', 'GeoLocation',
                      'Access', 'Backbone', 'Commercial', 'Customer', 
                      'Testbed', 'Transit']

        dw = csv.DictWriter(open('summary.csv','w'), delimiter=',', 
                            fieldnames=fieldnames, restval='')
        dw.writeheader()

        for node, data in sorted(summary_data.items()):
            node_dict = {}
            network = data['Network']
            node_dict['Network'] = network
            # All the other properties to include
            node_dict['Access'] = data['Access']
            node_dict['Backbone'] = data['Backbone']
            node_dict['Commercial'] = data['Commercial']
            node_dict['Customer'] = data['Customer']
            node_dict['Developed'] = data['Developed']
            node_dict['GeoExtent'] = data['GeoExtent']
            node_dict['GeoLocation'] = data['GeoLocation']
            node_dict['Layer'] = data['Layer']
            node_dict['Name'] = node
            node_dict['Network'] = data['Network']
            node_dict['Nodes'] = data['Nodes'] 
            node_dict['Edges'] = data['Edges'] 
            node_dict['Testbed'] = data['Testbed']
            node_dict['Transit'] = data['Transit']
            node_dict['Type'] = data['Type']

            if network_list.count(network) > 1:
                # Multiple observations of this network, record date
                node_dict['Date'] = data['NetworkDate']
            dw.writerow(node_dict)



        # write csv to output directory


    # Pre-processing for html and gallery as they both use these dictionaries
    if options.html:
        for network, data in summary_data.items():
            # Set any missing entries to blank
            for key in metadata_headings:
                if key not in data:
                    summary_data[network][key] = ""

    if options.gallery and options.map_dir:
        # See which networks have an associated map
        gallery_data = [] 
        for network in summary_data:
            map_file = "{0}{1}.jpg".format(options.map_dir, network)
            if os.path.isfile(map_file):
                # Map file  exists, add to gallery
                gallery_data.append(network)
            else:
                print("No plot file found for {0}".format(network))
            
    if options.html:
        #TODO: if matlab, plot, gml, graphml etc set, then provide hyperlink 
        date = strftime("%A %d %B %Y")

        lookup = TemplateLookup(directories=[ "templates" ],
                                module_directory="/tmp/mako_modules")       
        html_template = lookup.get_template("html.mako")
        f_html = open("{0}/dataset.html".format(output_path), "w")
        f_html.write(html_template.render(
            summary_data = summary_data,
            gallery_data = gallery_data,
            metadata_headings = metadata_headings,
            archive_file = archive_file,
            date = date,
        ))    
        f_html.close()
       
    if options.gallery:
        if options.map_dir:
            date = strftime("%A %d %B %Y")          
            #TODO: if matlab, plot, gml, graphml etc set, then provide hyperlink 
            # Set any missing entries to blank
            lookup = TemplateLookup(directories=[ "templates" ],
                                    module_directory="/tmp/mako_modules")       
            gallery_template = lookup.get_template("gallery.mako")
            f_html = open("{0}/gallery.html".format(output_path), "w")
            f_html.write(gallery_template.render(
                summary_data = summary_data,
                gallery_data = gallery_data,
                date = date,
            ))    
            f_html.close()
        else:
            print ("Please specify maps source directory to use in creating"
                   " gallery")


    if options.archive:
        archive_fh = zipfile.ZipFile(
            "{0}/{1}".format(output_path, archive_file), "w")
        for name in archive_filelist:
            archive_fh.write(name, os.path.basename(name), zipfile.ZIP_DEFLATED)
        archive_fh.close()
        

if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        pass    



#! /usr/bin/env python

from pyparsing import Word, alphas, nums, OneOrMore, Literal, Group, Optional
import networkx as nx
import optparse
import os
import glob
import re
import sys
import logging
import logging.handlers

#TODO: fix logging support

opt = optparse.OptionParser()
opt.add_option('--directory', '-d', help="process directory")
#TODO: make this required if using geocoding
opt.add_option('--geocode', '-g',  action="store_true",
                default=False,
                help="Attempt to lookup co-ordinates")
opt.add_option('--debug',  action="store_true",
                default=False,
                help="Print debugging information to terminal")
opt.add_option('--geonames_username', '-u', 
                default=False,
                help="Username for Geonames web service")
opt.add_option('--skip_cache',  action="store_true",
                default=False,
                help="Don't use cache for geocoding")
opt.add_option('--intra_as',  action="store_true",
                default=False,
               help="Only process Intra-AS topologies ie xxx:xxx not xxx:yyy "
               "folder format")

options = opt.parse_args()[0]

logger = logging.getLogger("rocketfuel2zoo")
logger.setLevel(logging.DEBUG)
formatter = logging.Formatter('%(levelname)-6s %(message)s')
ch = logging.StreamHandler()
ch.setLevel(logging.INFO)
if options.debug:
    ch.setLevel(logging.DEBUG)
ch.setFormatter(formatter)
logging.getLogger('').addHandler(ch)
log_dir = "."
LOG_FILENAME =  os.path.join(log_dir, "rocketfuel2zoo.log")
LOG_SIZE = 2097152 # 2 MB
fh = logging.handlers.RotatingFileHandler(
    LOG_FILENAME, maxBytes=LOG_SIZE, backupCount=5)
fh.setLevel(logging.DEBUG)
formatter = logging.Formatter("%(asctime)s %(levelname)s %(message)s")
fh.setFormatter(formatter)
logging.getLogger('').addHandler(fh)

def main():
    #TODO: make this a callback function for verification
    if options.geocode and not options.geonames_username:
        logger.error("Please enter a Geonames username to use geocoding")
        sys.exit(0)

    directory = options.directory + os.sep
    output_path = directory + os.sep + "rf2zoo" + os.sep
    output_path = os.path.normpath(output_path)
    if not os.path.isdir(output_path):
        os.mkdir(output_path)

    logger.info("Writing to %s" % output_path)
    all_folders = glob.glob(directory + "*")

    if options.geocode:
        from geocoder import Geocoder
        geocoder = Geocoder(options.geonames_username, options.skip_cache)

    for folder in all_folders:
        if not os.path.isdir(folder):
            continue
        #tODO: allow support for specifying the edge file directly

        # Check valid format of x:y where x, y are valid ASNs
        (_, data_folder) = os.path.split(folder)
        if not re.match("\d+:\d+", data_folder):
            continue

        # Only do internal topologies
        if options.intra_as:
            (asn_a, asn_b) = data_folder.split(":")
            if asn_a != asn_b:
                continue
        
        logger.info("Processing %s" % data_folder)

        G = nx.Graph()

        #TODO: check on bidirectionality
        #TODO: download file if not present
        #TODO: add pyparsing to topzootools dependencies for Pypi

        # Definitions for parsing
        colon = Literal(":").suppress()
        #TODO: see if comma built in
        comma = Literal(",").suppress()
        arrow = Literal("->").suppress()
        ASN = Word(nums)
        ObsCount = Word(nums)
        # Can have . eg St. Louis
        place_name = Group(
            OneOrMore(Word(alphas+".")) 
            + Optional(comma) +
            Optional(Word(alphas)))

        node = Group(ASN + colon + place_name)
        entry = (node + arrow + node + ObsCount)

        filename = folder + os.sep + "edges"
        f = open(filename)

        #ToDO; print unparsed lines

        #ToDo: push all into dict and use add_nodes_from and add_edges_from

        for line in f:
            #print line
            processed = entry.parseString(line)
            #print processed
            (src, dst, obs_count) = processed
            src_asn, src_place = src
            dst_asn, dst_place = dst
            src_place = " ".join(src_place)
            dst_place = " ".join(dst_place)
            # Use simple string for ID (not list returned from Pyparsing which
            # don't always hash the same
            src_id = "%s %s" % (src_place, src_asn)
            dst_id = "%s %s" % (dst_place, dst_asn)
            # use full hash for node id
            G.add_node(src_id, ASN=int(src_asn), label =src_place )
            G.add_node(dst_id, ASN=int(dst_asn), label=dst_place )
            G.add_edge(src_id, dst_id, obs_count = int(obs_count))

        f.close()

        # Relabel nodes to have id based on index

        if options.geocode:
                G = geocoder.geocode(G)
                geocoder.save_cache()

        
        G.name = data_folder
        G.graph['Creator'] = 'rocketfuel2zoo'

        out_file = output_path + os.sep + data_folder.replace(":", "_") + ".gml"

        #import pprint
        #pprint.pprint( sorted(G.nodes()))
        #pprint.pprint( G.nodes(data=True))
        nx.write_gml(G, out_file)

if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        pass



#! /usr/bin/env python

import networkx as nx

import os
import glob
import sys


#import numpy as np

import csv
import re
import optparse

import pprint as pp

from geocoder import Geocoder

#toDO; expand regex to capture LinkLabel "OC-192c"

#TODO: Move options and args parsing out of body of code, and make them pass
#params to main body of code which is a function - so can call programatically
#not just on command link And then can unit test the graphk
#TODO:  Also allow note on map like legend: rather than needig full csv
#TODO: could also store the SVN revision of the class.csv file

#TODO: add option to mark outliers

#TODO: write up a  test network containing differnt nodes and external nodes
# and hyperedges and edges labelled to check the output after the merger
# and converter (use this in documentation to explain the steps)

#TODO add comment to graph metadata if nodes skipped - similar to filename asdmn
# renaming base on args

#tODO: handle 2011-01-31 13:45:06,402 DEBUG AttributeError 'NoneType'
# object has no attribute 'encode'

#TODO: make merge by default only do classified nodes

#TODO: make consistent filename eg garr and kentman with dates

#ToDO: also allow country to be set in {{}} format
# and handle appropriately in cleanup

#TODO check case of nodes being left yellow and external nodes also being yellow
#being incorrectly tagged with the yellow internal node tag

#TODO: kill thread if get max exceeded geonames error
#TODO: auto warn about disconnected nodes
#TODO: add option to geocode external networks, only if {{}} included,
#otherwise skip if triangle
#TODO: Finish cleaning up merge script to remove sql
#TODO: allow option to skip cache - when manually add entries...
#and then append to cache/update?
#TODO: look at licence/credit for geonames
#don't geocode without it
#TODO: fix up iteration through places for region... eg each state seperately
#TODO: check tgat 2.5G and 500M get converted to edge data
#TODO: work out why #123456 fails with encode error first time
#TODO: make + behave with >


"""
Documentation:
make note in doc to not use capitals for names - as capitalized
and less than 5 characters are skipped
make note that merge and plot use cache, based on file modified,
so if rename files then won't be picked up... solution is to manually
delete cache file
write about how can use the map view of geonames to find places that match
Do examples of using zoo data in yed, gephi, matlab, matlab bgl, r library,
and networkx
tip : can often use state names to disambiguate - find well known place,
and then states from that
"""



import logging
import logging.handlers



#TODO fix bug wher eneed to append / to dir name for convert and merge
#TODO be careful not to apply colour legend classification to stop sign nodes
# which are used to represent hyperedges - but is ok for stop sign nodes
# with labels, which may just have doubted labels
#TODO catch errors in converter script, print, continue on to next network


opt = optparse.OptionParser()
opt.add_option('--file', '-f',
            help="Load data from FILE")

opt.add_option('--directory', '-d',
            help="process directory")


opt.add_option('--output_dir', '-o',
            help="process directory")

opt.add_option('--skip_cache',  action="store_true",
                default=False,
                help="Don't use cache for geocoding")

opt.add_option('--debug',  action="store_true",
                default=False,
                help="Print debugging information to terminal")

opt.add_option('--geocode', '-u', 
                default=False,
                help="Username for Geonames web service")

options = opt.parse_args()[0]

#TODO: work out best way to put in function but make globally accessible
logger = logging.getLogger("yed2zoo")
logger.setLevel(logging.DEBUG)
formatter = logging.Formatter('%(levelname)-6s %(message)s')
ch = logging.StreamHandler()
ch.setLevel(logging.INFO)
if options.debug:
    ch.setLevel(logging.DEBUG)
ch.setFormatter(formatter)
logging.getLogger('').addHandler(ch)
log_dir = "."
LOG_FILENAME =  os.path.join(log_dir, "yed2zoo.log")
LOG_SIZE = 2097152 # 2 MB
fh = logging.handlers.RotatingFileHandler(
    LOG_FILENAME, maxBytes=LOG_SIZE, backupCount=5)
fh.setLevel(logging.DEBUG)
formatter = logging.Formatter("%(asctime)s %(levelname)s %(message)s")
fh.setFormatter(formatter)
logging.getLogger('').addHandler(fh)

def convert_to_undirected(graph):
    if graph.is_multigraph():
        """
        When NetworkX converts from MultiDiGraph to MultiGraph, it will discard
        an edge if it exists in both directions, eg (1, 2) and (2, 1) will
        discard (2, 1). For maps traced in yED, the existence of a link means it
        was deliberately added, and we wish to keep it. In above example we wish
        to have 2 edges between nodes 1 and 2. To do this, we need to manually
        add another link from (1, 2) and remove (2, 1)"""
        for src, dst, key, data in graph.edges(data=True, keys=True):
            # See if reverse link
            if graph.has_edge(dst, src):
                # Reverse link exists, want all to go same direction
                # All edges for node pair need to originate from same node,
                # choose lower node id to decide
                if src > dst:
                    # Switch the link
                    graph.add_edge(dst, src, attr_dict=data)
                    graph.remove_edge(src, dst, key)

    # Undirected version of graph (both single and multi edge)
    graph = graph.to_undirected()

    return nx.MultiGraph(graph)


def extract_speed(data, label):
    speed_multipliers = {
        'K': 1e3,
        'M': 1e6,
        'G': 1e9,
    }
    num_units = "(\d+\.*\d*)\s*(k|K|m|M|g|G|t|T)(b?p?/?s?)"
    speed_invalidators = "[>|<|<=|>=|=>|=<|*|x|-]"
    if extract_speed:
        if re.search( speed_invalidators+"\s*"+num_units, label, re.IGNORECASE):
            # eg >45Mbps, <=1gbps, 2x>2.5G, etc
            # Ambiguous as to actual speed -> skip
            pass
        elif len(re.findall(num_units, label, re.IGNORECASE)) > 1:
            # Speed should only occur once
            pass
        else:
            # Try and extract units if present
            re_m = re.search(num_units, label, re.IGNORECASE)
            if re_m:
                speed = re_m.group(1)
                units = re_m.group(2)
                # Get the bps bp/s bit as well to remove from note
                units_extra = re_m.group(3)
                data['LinkNote'] = data['LinkNote'].replace(speed, "")
                data['LinkNote'] = data['LinkNote'].replace(units, "")
                data['LinkNote'] = data['LinkNote'].replace(units_extra, "")
                data['LinkSpeed'] = speed
                units = units.upper()
                data['LinkSpeedUnits'] = units
                data['LinkSpeedRaw'] = speed_multipliers[units] * float(speed)

    return data

def extract_type(data, label):
    edge_types = "OC-\d+|T\d|E\d|DS-\d+|STM-\d+|SDH|Fib[e|r]{2}|Optical|GIGE|FE"
    edge_types += "|Ethernet|Satellite|Wireless|Microwave|Serial|ATM|MPLS"

    if len(re.findall(edge_types, label, re.IGNORECASE)) > 1:
        # Should only occur once -> ignore
        # see if occurs as type (over|via) type,
        # eg Ethernet over SDH, Ethernet over Satellite, etc
        over_via =  r"({0})\s(over|via)\s({0})".format(edge_types)
        re_m = re.search(over_via, label, re.IGNORECASE)
        if re_m:
            temp = re_m.group(0)
            data['LinkNote'] = data['LinkNote'].replace(temp, "")
            data['LinkType'] = temp
        else:
            # skip this entry as invalid
            pass
    else:
        re_m = re.search(edge_types, label, re.IGNORECASE)
        if re_m:
            temp = re_m.group(0)
            data['LinkNote'] = data['LinkNote'].replace(temp, "")
            # pretty up the type
            # eg ethernet -> Ethernet, ETHERNET -> Ethernet, STM -> STM ie same
            if not temp.isupper():
                # Not upper case, tidy up
                # Ignore all upper case, eg avoid ATM -> Atm, SDH -> Sdh
                temp = temp.title()
            data['LinkType'] = temp
    return data

def extract_temporal(data, label):
    temporal = "Future|Planned|Current|Proposed"
    temporal += "|Planning|Under Construction|Under Development"
    re_m = re.search(temporal, label, re.IGNORECASE)
    if re_m:
        temp = re_m.group(0)
        data['LinkNote'] = data['LinkNote'].replace(temp, "")
        data['LinkStatus'] = temp.title()
    return data

def extract_edge_data(label):

    #TODO: also do the raw speed for edges (optional command line  argument
    # eg 10M is 10 * 10^6 = 10,000,000 in the GML

    # The "Note" starts off as label and is trimmed as data is extracted
    data = {'LinkLabel': label, 'LinkNote': str(label), 'LinkSpeedUnits': '',
            'LinkType': '', 'LinkSpeed': '', 'time': ''}

    do_extract_speed = True

    # Note: Could likely optimise the following into a single regexp
    # Skip speed if ambiguities, ie if - or + present and not in STM- or OC-
    if re.search(r"-+", label):
        # Contains + or -
        # Remove OC- and STM- and check if still contains -
        temp = re.sub(r"OC-\d+|STM-\d+", "", label)
        if re.search(r"-+", temp):
            # label contains - or + not inside OC- or STM-, so skip speed
            do_extract_speed = False

    if do_extract_speed:
        data = extract_speed(data, label)
    data = extract_type(data, label)
    data = extract_temporal(data, label)

    # Tidying up: remove note if it is just punctuation
    # eg () or + left over from extraction
    if ( (len(data['LinkNote']) > 0) and
        (len(re.findall("\w+", data['LinkNote'], re.IGNORECASE)) == 0) ):
        # Non-trivial length and no alphanumeric characters present so wipe note
        data['LinkNote'] = ''

    # Remove any empty fields
    for key, val in data.items():
        if val == '':
            del data[key]

    # Don't include note if same as label - redundant
    if ('LinkNote' in data) and (data['LinkNote'] == data['LinkLabel']):
        del data['LinkNote']

    return data


def extract_svn_version(path):
    # Tries to obtain SVN version from path
    # http://code.djangoproject.com/browser/django/trunk/django/utils/version.py

    if os.path.isfile(path):
        # likely a single file passed in
        path = os.path.split(path)[0]

    entries_file = "{0}/.svn/entries".format(path)
    # check exists
    if not os.path.exists(entries_file):
        return None
    f_entries = open(entries_file, 'r')
    # line after 'dir' is the revision number
    next_line = False
    for line in f_entries:
        if next_line:
            # this line has the version number
            return int(line)
        if line.strip() == 'dir':
            next_line = True
    f_entries.close()

def main():
    network_files = []
    if options.file:
        network_files.append(options.file)

    if options.directory:
        network_files = glob.glob(options.directory + "*.graphml")

    if len(network_files) == 0:
        logger.warn("No files found. Specify -f file or -d directory")
        sys.exit(0)

    if options.directory:
        path = options.directory
    elif options.file:
        path = os.path.split(options.file)[0]
        # Get full path - don't want to create in root dir
        path = os.path.abspath(path)

    if options.output_dir:
        output_path = options.output_dir
    else:
        output_path = path + os.sep + "zoogml"

    if options.geocode:
        output_path += "_geocoded"

    if not os.path.isdir(output_path):
        os.mkdir(output_path)

    # clean up path
    output_path = os.path.normpath(output_path)

    #TODO: check nx write pickle uses cPickle

    #TODO: make this optional
    output_pickle_path = output_path + os.sep + "cache"       
    if not os.path.isdir(output_pickle_path):
        os.mkdir(output_pickle_path)

    geocoder = Geocoder(options.geocode, options.skip_cache)

    #output_path += strftime("%Y%m%d_%H%M%S")
    logger.info("Saving to folder: %s" % output_path)
    # and create cache directory for pickle files
    pickle_dir = path + os.sep + "cache"       
    if not os.path.isdir(pickle_dir):
        os.mkdir(pickle_dir)

        #ToDO: check why sorting
    for source_file in sorted(network_files):
        # Extract name of network from file path
        filename = os.path.split(source_file)[1]
        net_name = os.path.splitext(filename)[0]
        logger.info( "Converting {0}".format(net_name))

        pickle_file = "{0}/{1}.pickle".format(pickle_dir, net_name)
        if (os.path.isfile(pickle_file) and
            os.stat(source_file).st_mtime < os.stat(pickle_file).st_mtime):
            # Pickle file exists, and source_file is older
            graph = nx.read_gpickle(pickle_file)
        else:
            # No pickle file, or is outdated
            graph = nx.read_graphml(source_file)
            nx.write_gpickle(graph, pickle_file)

        graph = convert_to_undirected(graph)
        # Check for self loops
        for n, data in graph.nodes(data=True):
            if n in graph.neighbors(n):
                logger.warn( "Self loop {0} {1}".format(data['label'], n))

        # if all nodes in network are internal, remove redundant internal tag
        # this makes checking if internal too complex, keep tag for all nodes
        """
        if all(data['Internal'] == 1 for n,data in graph.nodes(data=True)):
            # all internal, remove tag
            for n in graph.nodes():
                del graph.node[n]['Internal']
        """

        #*********************************
        # Geocoding (optional)
        #*********************************
        if options.geocode:
            graph = geocoder.geocode(graph)
            geocoder.save_cache()

        #*********************************
        # Remove graphics data
        #*********************************

        # Will be overwritten with name from metadata if present
        network_label = net_name

        # extract edge data
        for src, dst, key, data in graph.edges(data=True, keys=True):
            if 'label' in data:
                label = data['label']
                extracted_data = extract_edge_data(label)
                # Replace the edge data with extracted
                graph.edge[src][dst][key] = extracted_data

        # remove empty note
        if 'Note' in graph.graph and graph.graph['Note'] == '':
            del graph.graph['Note']

        # Strip & as yEd fails on it
        #TODO: use html entitites fn for this
        network_label_clean = network_label.replace("&", "and")

        # Set other graph attributes
        graph.graph['Creator'] = "Topology Zoo Toolset"
        graph.graph['Version'] = "1.0"

        graph.graph['label'] = network_label_clean

        #*********************************
        #OUTPUT - Write the graphs to files
        #*********************************

        filename = network_label.lower().strip()
        filename = filename.title()
        filename = filename.replace(" ", "_")

        pattern = re.compile('[\W_]+')
        filename = pattern.sub('', filename)
        # And also the pickle cache - write first so older timestamp
        pickle_out_file =  "{0}/{1}.pickle".format(output_pickle_path, filename)
        nx.write_gpickle(graph, pickle_out_file)

        gml_file =  "{0}/{1}.gml".format(output_path, filename)
        nx.write_gml(graph, gml_file)
        logger.info("Wrote to %s" % gml_file)


#TODO: move the nested functions out to be main functions - and pass in
 #appropriate args (setup args in prev line to nested call)

if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        pass



#!/usr/bin/env python
from setuptools import setup, find_packages

setup (
     name = "topzootools",
     version = "0.3.13",
     description = "Processing and conversion tools for Internet Topology Zoo",
     long_description="Processing and conversion tools for Internet Topology Zoo",
               
     # simple to run 
     entry_points = {
         'console_scripts': [
             'yed2zoo = TopZooTools.yed2zoo:main',    
             'zooconvert = TopZooTools.converter:main',         
             'zooplot = TopZooTools.geoplot:main',
             'zoofilter = TopZooTools.filter:main',
             'zoostats= TopZooTools.zoostats:main',
             'zoogeocode= TopZooTools.geocoder:main',
             'rocketfuel2zoo= TopZooTools.rocketfuel2zoo:main',
         ],
     },

     author='Simon Knight',
     author_email="simon.knight@adelaide.edu.au",
     url="http://www.topology-zoo.org",
     packages=['TopZooTools'],  
    
        #TODO: update this package data to reflect all scripts called
     package_data = {'': ['converter.py', 'geoplot.py', 'yed2zoo.py',
                          'rocketfuel2zoo.py', 'filter.py', 'geocoder.py']},
     
     install_requires=['mako', 'networkx', 'matplotlib', 'numpy', 'pyparsing'],
     classifiers = [
         "Programming Language :: Python",
         "Development Status :: 3 - Alpha",
         "Intended Audience :: Science/Research",
         "Intended Audience :: System Administrators",
         "Intended Audience :: Telecommunications Industry",
         "License :: Other/Proprietary License",
         "Operating System :: MacOS :: MacOS X",
         "Operating System :: POSIX :: Linux",
         "Topic :: System :: Networking",
         "Topic :: System :: Software Distribution",
         "Topic :: Scientific/Engineering :: Mathematics",
         ],     
 
)



#! /usr/bin/env python

import networkx as nx

import os
import glob
import sys
from time import strftime

import optparse

import pprint

def category_counts(data):
    keys = set(data)
    for key in sorted(keys):
        print "{0}\t{1}".format(key.ljust(10), data.count(key))
    print "{0}\t{1}".format("Total".ljust(10), len(data))


def main():
    opt = optparse.OptionParser()
    opt.add_option('--file', '-f', help="Load data from FILE")
    opt.add_option('--directory', '-d', help="process directory")
    opt.add_option('--output_dir', '-o', help="process directory")
    options = opt.parse_args()[0]

    network_files = []
    if options.file:
        network_files.append(options.file)

    if options.directory:
        network_files = glob.glob(options.directory + "*.gml")

    if len(network_files) == 0:
        print "No files found. Please specify -f file or -d directory"
        sys.exit(0)

    if options.directory:
        path = options.directory
    elif options.file:
        path, filename = os.path.split(options.file)

    pickle_dir = path + "/cache"
    if not os.path.isdir(pickle_dir):
        os.mkdir(pickle_dir)

    summary_data = {}

    disconnected_networks = {}

    for net_file in network_files:

        # Extract name of network from file path
        path, filename = os.path.split(net_file)
        network_name = os.path.splitext(filename)[0]

        #print "Converting: {0}".format(network_name)

        #if network_files.index(net_file) > 20:
        #    break

        pickle_file = "{0}/{1}.pickle".format(pickle_dir, network_name)
        if (os.path.isfile(pickle_file) and
            os.stat(net_file).st_mtime < os.stat(pickle_file).st_mtime):
            # Pickle file exists, and source_file is older
            graph = nx.read_gpickle(pickle_file)
        else:
            print "Caching copy of %s" % network_name
            # No pickle file, or is outdated
            graph = nx.read_gml(net_file)
            nx.write_gpickle(graph, pickle_file)

        graph = graph.to_undirected()
        graph = nx.MultiGraph(graph)


        external_nodes = [ n for n in graph.nodes()
                          if 'Internal' in graph.node[n] and
                          graph.node[n]['Internal'] == 0]
        graph.remove_nodes_from(external_nodes)
        hyperedge_nodes = [ n for n in graph.nodes()
                          if 'hyperedge' in graph.node[n] and
                          graph.node[n]['hyperedge'] == 1]

        if not nx.is_connected(graph):
            components = [component for component in
                          nx.connected_components(graph)]
            component_sizes = []
            for component in components:
                # Only want internal, non hyperedge nodes
                component_size = len([n for n in component 
                                      if (n not in external_nodes 
                                          and n not in hyperedge_nodes)]) 
                component_sizes.append(component_size)

            disconnected_networks[network_name] = component_sizes

        summary_data[network_name] = {} 

        for key, val in graph.graph.items():
            # And also store in html data
            summary_data[network_name][key] = val

    # And range
    print "------"
    print "%i networks" % len(summary_data)


    # And range
    print "------"
    types_keys = ['Access', 'Backbone', 'Customer', 'Testbed', 'Transit', 'IX']
    print "Type".ljust(10) + "\tCOM\tREN\tTotal"
    for key in types_keys:
        # All networks that have this statistic
        all_nets = [network for network, data in summary_data.items() if 
                    key in data and data[key] == 1]
        # Commercial subset
        com_nets = [network for network in all_nets 
               if summary_data[network]['Type'] == 'COM']
        # Research subset
        ren_nets = [network for network in all_nets 
               if summary_data[network]['Type'] == 'REN']

        print "{0}\t{1}\t{2}\t{3}".format(key.ljust(10), len(com_nets), 
                                          len(ren_nets), len(all_nets))

    # And range
    print "------"
    print "Range: "
    all_vals = [d["GeoExtent"] for d in summary_data.values()]
    category_counts(all_vals)

    print "-----"
    print "Connected Components:"
    print "Network".ljust(10) + "\tConnected Components"
    for key, val in sorted(disconnected_networks.items()):
        print "%s\t%s" % (key.ljust(10), "/".join([str(v) for v in val]))

    # Layer
    print "----------"
    print "Layer:"
    all_vals = [d["Layer"] for d in summary_data.values()]
    category_counts(all_vals)

    #print "No layer: " + ",".join([network for network, data
    #                               in summary_data.items()
    #                               if data["Layer"] == ""])

    # Com and Ren
    print "----------"
    print "Type:"
    all_vals = [d["Type"] for d in summary_data.values()]
    category_counts(all_vals)


    print "----------"
    print "Provenance:"
    all_vals = [d["Provenance"] for d in summary_data.values()]
    category_counts(all_vals)

    # and dates
    print "----------"
    print "Date:"
    # We need to look at current/historical/dynamic in conjunction with the date
    all_dates = []
    # The value depends on the DateType field. Historical and current networks
    # are classified by year. Dynamic networks are kept as dynamic.
    all_dates += [data['NetworkDate'] for data in summary_data.values()
                  if data['DateType'] == "Historic"]
    #all_dates += [data['NetworkDate'] for data in summary_data.values()
    #              if data['DateType'] == "Current"]
    all_dates += ['Current' for data in summary_data.values()
                  if data['DateType'] == "Current"]

    current_nets = [network for network, data in summary_data.items()
                  if data['DateType'] == "Current"]
    #for net in current_nets:
    #    print ("%s %s" % (net, summary_data[net]['DateObtained']))
    # Truncate date from YYYY-MM to just YYYY
    # Get list of dates first, then iterate using it (don't want to modify list
    # while iterating over it)
    yyyy_mm = [date for date in all_dates if len(date) == 7 and date[4] == "-"
               and date[:4].isdigit()]
    for date in yyyy_mm:
        all_dates.remove(date)
        all_dates.append(date[:4])

    #all_dates = [date.split("-")[0] for date in all_dates 
    #             if len(date) == 7 and date[5] == "-"]
    # Dynamic networks remain as dynamic
    all_dates += ['Dynamic' for data in summary_data.values()
                  if data['DateType'] == "Dynamic"]
    category_counts(all_dates)
   
    """
    histogram = []
    numeric_dates = [int(date) for date in all_dates if date.isdigit()]
    bins = [min(numeric_dates), 2000, 2005, 2010]
    for date in numeric_dates:
        for bin in bins:
            if date <= bin:
                histogram[bin]
    """

if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        pass    
